//-----------------------------------------------------------------------------
//  
// linux/drivers/video/epson/s1d13xxxfb_2D.c --  
// 2D Hardware accecerated functions for Epson S1D13xxx controller frame 
// buffer drivers. 
// 
// Copyright(c) Seiko Epson Corporation 2000-2005.
// All rights reserved.
//
// This file is subject to the terms and conditions of the GNU General Public
// License. See the file COPYING in the main directory of this archive for
// more details.
//
//---------------------------------------------------------------------------- 
//---------------------------------------------------------------------------- 
//
// WARNING: This file uses a chip specific include file. The include file 
// contains all hardware information needed to generate proper timings and 
// access registers
//
// This driver supports one of the following Epson CRT/TV/LCD controllers:
//
// S1D13A05
//
//-----------------------------------------------------------------------------
#include <linux/kernel.h>
#include <linux/fb.h>

#include "s1d13xxxfb.h"

#ifdef CONFIG_FBCON_EPSON_2D

//-----------------------------------------------------------------------------
// Local definitions
//----------------------------------------------------------------------------- 
#define STRIDE S1D_DISPLAY_SCANLINE_BYTES

// BitBlit engine register address
#define REG_BLT_BASE                        0x8000
#define REG8000_BLT_CONTROL                 0x8000
#define REG8004_BLT_STATUS                  0x8004
#define REG8008_BLT_COMMAND_ROP             0x8008
#define REG800C_BLT_SOURCE_START_ADDR       0x800C
#define REG8010_BLT_DEST_START_ADDR         0x8010
#define REG8014_BLT_MEMORY_ADDR_OFFSET      0x8014
#define REG8018_BLT_WIDTH                   0x8018
#define REG801C_BLT_HEIGHT                  0x801C
#define REG8020_BLT_BACKGROUND_COLOR        0x8020
#define REG8024_BLT_FOREGROUND_COLOR        0x8024

// to prevent endless looping
#define LOOP_LIMIT_2D                                      10000

// While the FIFO is being written to by the CPU, it is also being emptied by the S1D.
// If the S1D empties the FIFO faster than the CPU can fill it, it may not be possible to
// cause an overflow/underflow. In these cases, performance can be improved by not
// monitoring the FIFO status. However, this is very much platform dependent and must be
// determined for each system.
#define BLT_TEST_FIFO


//-----------------------------------------------------------------------------
// Local globals
//----------------------------------------------------------------------------- 
static volatile u16* BltDataAddr;     // Address of BitBlit data register
static u32 BitBLTCtrlReg;      // Contents of BitBlit control register

//-----------------------------------------------------------------------------
// Function Prototypes 
//-----------------------------------------------------------------------------
void s1d13xxxfb_2DEngineInit( volatile u16* bltDataAddr, struct fb_info *info, struct fb_ops *ops );

void s1d13xxxfb_fillrect (struct fb_info *info, const struct fb_fillrect *rect);
void s1d13xxxfb_copyarea (struct fb_info *info, const struct fb_copyarea *region);
void s1d13xxxfb_imageblit (struct fb_info *info, const struct fb_image *image);
int  s1d13xxxfb_sync (struct fb_info *info );


//-----------------------------------------------------------------------------
// Private Function Prototypes 
//-----------------------------------------------------------------------------
static int  s1d13xxxfb_IsSupported (u32 bpp);
static int  s1d13xxxfb_WaitForBltEnd (void);
static int  s1d13xxxfb_RectOverlap (const struct fb_copyarea *region);
static void s1d13xxxfb_colorblit (struct fb_info *info, const struct fb_image *image,
                                  u32 bpp, int nTotalWords, volatile u16* dst);
static void s1d13xxxfb_lowblit (struct fb_info *info, const struct fb_image *image,
                                u32 bpp, int nTotalWords, volatile u16* dst);

 
//-----------------------------------------------------------------------------
// s1d13xxxfb_2DEngineInit()
//-----------------------------------------------------------------------------
//
// Purpose: Initializes the 2D engine
//
//-----------------------------------------------------------------------------
void s1d13xxxfb_2DEngineInit( volatile u16* bltDataAddr, struct fb_info *info, struct fb_ops *ops )
{
	// Pointer to BlitBit Data Register
	BltDataAddr = bltDataAddr;

	// Fill in flags for which HW accelerated features are supported
	info->flags |= FBINFO_HWACCEL_COPYAREA | FBINFO_HWACCEL_FILLRECT | FBINFO_HWACCEL_IMAGEBLIT;

	// Set system pointer to HW accelerated functions
	ops->fb_fillrect    = s1d13xxxfb_fillrect,
	ops->fb_copyarea    = s1d13xxxfb_copyarea,
	ops->fb_imageblit   = s1d13xxxfb_imageblit,
	ops->fb_sync        = s1d13xxxfb_sync;

	// Set Control Register
	switch( S1D_DISPLAY_BPP ) {
	case 8:
		BitBLTCtrlReg = 0x00000000;
		s1d13xxxfb_write_reg(REG8000_BLT_CONTROL, BitBLTCtrlReg);
		break;
	case 16:
		BitBLTCtrlReg = 0x00040000;
		s1d13xxxfb_write_reg(REG8000_BLT_CONTROL, BitBLTCtrlReg);
		break;
	}
}

//-----------------------------------------------------------------------------
// s1d13xxxfb_IsSupported()
//-----------------------------------------------------------------------------
//
// Purpose: Checks if the 2D engine supports the color depth
//
//-----------------------------------------------------------------------------
static int s1d13xxxfb_IsSupported( u32 bpp )
{
#ifdef CONFIG_FBCON_EPSON_S1D13A05
	switch( bpp ) {
	case 8:
	case 16:
		return 1;
	default:
		return 0;
	}
#else
	return 0;
#endif
}

//-----------------------------------------------------------------------------
// s1d13xxxfb_WaitForBltEnd()
//-----------------------------------------------------------------------------
//
// Purpose: Waits for 2D (BitBLT) engine to complete operation
//
//-----------------------------------------------------------------------------
static int s1d13xxxfb_WaitForBltEnd( void )
{
	u16 timeout = LOOP_LIMIT_2D;

	// wait until the Blt Engine is idle and Command Queue is empty
	// or timeout limit reached
	while ( (s1d13xxxfb_read_reg(REG8004_BLT_STATUS) & 0x0001) && timeout--) {}

	if (timeout)
		return 1;
	else
		return 0;
}

//-----------------------------------------------------------------------------
// s1d13xxxfb_sync()
//-----------------------------------------------------------------------------
//
// Purpose: Waits for 2D (BitBLT) engine to complete operation and be idle
//
//-----------------------------------------------------------------------------
int s1d13xxxfb_sync( struct fb_info *info )
{
	// wait until the Blt Engine is idle and Command Queue is empty
	while (s1d13xxxfb_read_reg(REG8004_BLT_STATUS) & 0x0001) {}

	return 0;
}

//-----------------------------------------------------------------------------
// s1d13xxxfb_RectOverlap()
//-----------------------------------------------------------------------------
//
// Purpose: Verify if Source and Destination rectangles overlap.
//
//-----------------------------------------------------------------------------
static int s1d13xxxfb_RectOverlap( const struct fb_copyarea *region )
{
	int Sx0,Sx1,Dx0,Dx1,Sy0,Sy1,Dy0,Dy1;

	Sx0 = region->sx;
	Sx1 = Sx0 + region->width;
	Dx0 = region->dx;
	Dx1 = Dx0 + region->width;

	Sy0 = region->sy;
	Sy1 = Sy0 + region->height;
	Dy0 = region->sy;
	Dy1 = Dy0 + region->height;

	if (Dx0 > Sx1 || Dy0 > Sy1 || Dx1 < Sx0 || Dy1 < Sy0)
		return 0;

	return 1;
}

//-----------------------------------------------------------------------------
// s1d13xxxfb_fillrect()
//-----------------------------------------------------------------------------
//
// Purpose: Draws a rectangle using BitBlit engine
//
// Notes: 13A05 only supports rop operation ROP_COPY
//
//-----------------------------------------------------------------------------

void s1d13xxxfb_fillrect( struct fb_info *info, const struct fb_fillrect *rect )
{
	u32 dstAddr;
	u32 fg;
	u32 bytesPerPixel = info->var.bits_per_pixel / 8;

	// check if this feature is supported at current settings
	if( !s1d13xxxfb_IsSupported(info->var.bits_per_pixel) || (rect->rop != ROP_COPY) )
		return cfb_fillrect( info, rect );

	if (rect->width == 0 || rect->height == 0)
		return;

	// Wait for any BitBlit operations to finish
	if ( !s1d13xxxfb_WaitForBltEnd() )
		return cfb_fillrect( info, rect );

	// set Fill Blt
	s1d13xxxfb_write_reg(REG8008_BLT_COMMAND_ROP, 0x000C);

	s1d13xxxfb_write_reg(REG8014_BLT_MEMORY_ADDR_OFFSET, STRIDE/2);

	// Calculate screen address
	dstAddr = rect->dx*bytesPerPixel + rect->dy * STRIDE;
	s1d13xxxfb_write_reg(REG8010_BLT_DEST_START_ADDR, dstAddr);

	// Set height and width registers
	s1d13xxxfb_write_reg(REG8018_BLT_WIDTH, rect->width-1);
	s1d13xxxfb_write_reg(REG801C_BLT_HEIGHT, rect->height-1);

	// Determine color depending on color depth
	if (info->fix.visual == FB_VISUAL_TRUECOLOR || info->fix.visual == FB_VISUAL_DIRECTCOLOR )
		fg = ((u32 *) (info->pseudo_palette))[rect->color];
	else
		fg = rect->color;
	s1d13xxxfb_write_reg(REG8024_BLT_FOREGROUND_COLOR, fg);

	// Engage the blt engine.
	s1d13xxxfb_write_reg(REG8000_BLT_CONTROL, BitBLTCtrlReg | 0x00000001);
}

//-----------------------------------------------------------------------------
// s1d13xxxfb_copyarea()
//-----------------------------------------------------------------------------
//
// Purpose: Copy data from area to another using BitBlit engine
//
//-----------------------------------------------------------------------------
void s1d13xxxfb_copyarea( struct fb_info *info, const struct fb_copyarea *region )
{
	int direction;
	u32 dstAddr,srcAddr;
	u32 bytesPerPixel = info->var.bits_per_pixel / 8;

	// check if this feature is supported at current settings
	if( !s1d13xxxfb_IsSupported(info->var.bits_per_pixel) )
		return cfb_copyarea( info, region );

	if (region->width == 0 || region->height == 0)
		return;

	// estabilish a known src & dst. background
	srcAddr = region->sx * bytesPerPixel + region->sy * STRIDE;
	dstAddr = region->dx * bytesPerPixel + region->dy * STRIDE;

	// If the blit is not transparent, determine the direction of the blt.
	// Transparent blt supports only positive direction.
	direction = 0;

	if (s1d13xxxfb_RectOverlap(region)) {
	        if (dstAddr > srcAddr) {
			direction = 1;
			// negative direction : get the coords of lower right corner
			srcAddr += STRIDE *(region->height-1) + bytesPerPixel *(region->width-1);
			dstAddr += STRIDE *(region->height-1) + bytesPerPixel *(region->width-1);
		}
	}

	// wait for any pending blit to end
	if ( !s1d13xxxfb_WaitForBltEnd() )
		return cfb_copyarea( info, region );

	// program the BLT stride
	s1d13xxxfb_write_reg(REG8014_BLT_MEMORY_ADDR_OFFSET, STRIDE/2);

	// Set the Source addr
	s1d13xxxfb_write_reg(REG800C_BLT_SOURCE_START_ADDR, srcAddr);

	// Set the Destination addr
	s1d13xxxfb_write_reg(REG8010_BLT_DEST_START_ADDR, dstAddr);

	s1d13xxxfb_write_reg(REG8018_BLT_WIDTH, region->width-1);
	s1d13xxxfb_write_reg(REG801C_BLT_HEIGHT, region->height-1);

	// set the blt type
	if (direction)
		s1d13xxxfb_write_reg(REG8008_BLT_COMMAND_ROP, 0x000C0003);  // negative direction
	else
		s1d13xxxfb_write_reg(REG8008_BLT_COMMAND_ROP, 0x000C0002);  // positive direction

	// Engage the blt engine.
	s1d13xxxfb_write_reg(REG8000_BLT_CONTROL, BitBLTCtrlReg | 0x00000001);
}

//-----------------------------------------------------------------------------
// s1d13xxxfb_colorblit()
//-----------------------------------------------------------------------------
//
// Purpose: Transfers the image data to BitBlit engine for fb_imageblit ops for
//          higher depths
//
//-----------------------------------------------------------------------------
static void s1d13xxxfb_colorblit( struct fb_info *info, const struct fb_image *image,
                                  u32 bpp, int nTotalWords, volatile u16* dst )
{
#ifdef BLT_TEST_FIFO
	int j,nFIFO,byte8004;
#endif
	int i = 0;
	u16 pixel1, pixel2;

	while (nTotalWords > 0) {
#ifdef BLT_TEST_FIFO
		// read the FIFO status
		byte8004 = s1d13xxxfb_read_reg(REG8004_BLT_STATUS);
		if ((byte8004 & 0x30) == 0x20)      //FIFO at least half full, not full
			nFIFO = 1;
		else if ((byte8004 & 0x40) == 0)    //FIFO empty
			nFIFO = 16;
		else
			continue;
		for (j = 0; j < nFIFO && nTotalWords > 0; j++,nTotalWords--)
#else
		nTotalWords--;
#endif
		{
			// Uncompressed the bitmap image and transfer it to the 
			// BitBlit data reg space in word size chunks
			switch (bpp) {
			case 8:
				if (info->fix.visual == FB_VISUAL_TRUECOLOR || info->fix.visual == FB_VISUAL_DIRECTCOLOR) {
					pixel1 = (u16)((u32*)(info->pseudo_palette))[(u8)image->data[i++]];
					pixel2 = (u16)((u32*)(info->pseudo_palette))[(u8)image->data[i++]];
				} else {
					pixel1 = (u16)image->data[i++];
					pixel2 = (u16)image->data[i++];
				}
                    
				// 13A05 requires first byte goes into low byte of a word
				*dst = pixel2<<8 | (pixel1&0x00FF);
				break;
			case 16:
				if (info->fix.visual == FB_VISUAL_TRUECOLOR || info->fix.visual == FB_VISUAL_DIRECTCOLOR)
					*dst = (u16)((u32*)(info->pseudo_palette))[(u8)image->data[i++]];
				else
					*dst = (u16)image->data[i++];
				break;
			}
		}
	}
}

//-----------------------------------------------------------------------------
// s1d13xxxfb_lowblit()
//-----------------------------------------------------------------------------
//
// Purpose: Transfers the image data to BitBlit engine for fb_imageblit ops for
//          depth of 1.
//

//-----------------------------------------------------------------------------
static void s1d13xxxfb_lowblit( struct fb_info *info, const struct fb_image *image,
                                u32 bpp, int nTotalWords, volatile u16* dst )
{
#ifdef BLT_TEST_FIFO
	int j,nFIFO,byte8004;
#endif
	int i = 0, l = 8;
	u16 pixel1, pixel2;
	u32 fg, bg;

	if (info->fix.visual == FB_VISUAL_TRUECOLOR || info->fix.visual == FB_VISUAL_DIRECTCOLOR) {
		fg = ((u32*)(info->pseudo_palette))[image->fg_color];
		bg = ((u32*)(info->pseudo_palette))[image->bg_color];
	} else {
		fg = image->fg_color;
		bg = image->bg_color;
	}

	while (nTotalWords > 0) {
#ifdef BLT_TEST_FIFO
		// read the FIFO status
		byte8004 = s1d13xxxfb_read_reg(REG8004_BLT_STATUS);
		if ((byte8004 & 0x30) == 0x20)      //FIFO at least half full, not full
			nFIFO = 1;
		else if ((byte8004 & 0x40) == 0)    //FIFO empty
			nFIFO = 16;
		else
			continue;
		for (j = 0; j < nFIFO && nTotalWords > 0; j++,nTotalWords--)
#else 
		nTotalWords--;
#endif
		{
			// Uncompressed the bitmap image and transfer it to the 
			// BitBlit data reg space in word size chunks
			switch (bpp) {
			case 8:
				l--;
				if (image->data[i] & (1 << l))
					pixel1 = (u16)fg;
				else
					pixel1 = (u16)bg;

				if (!l) {l = 8; i++;}
				l--;
				if (image->data[i] & (1 << l))
					pixel2 = (u16)fg;
				else
					pixel2 = (u16)bg;
                    
				if (!l) {l = 8; i++;}

				// 13A05 requires first byte goes into low byte of a word
				*dst = pixel2<<8 | (pixel1&0x00FF);
				break;
			case 16:
				l--;
				if (image->data[i] & (1 << l))
					*dst = (u16)fg;
				else
					*dst = (u16)bg;

				if (!l) {l = 8; i++;}
				break;
			}
		}
	}
}

//-----------------------------------------------------------------------------
// s1d13xxxfb_imageblit()
//-----------------------------------------------------------------------------
//
// Purpose: Draws a image to the display using BitBlit engine
//
//-----------------------------------------------------------------------------
void s1d13xxxfb_imageblit( struct fb_info *info, const struct fb_image *image )
{
	u32 dstAddr;
	int nTotalWords = 0;
	/*int phase;*/
	volatile u16* dst;
	u32 bpp = info->var.bits_per_pixel;

	u32 width = image->width, height = image->height; 
	u32 dx = image->dx, dy = image->dy;
	int x2, y2, vxres, vyres;

	// Check if this feature is supported at current settings
	if (!s1d13xxxfb_IsSupported(info->var.bits_per_pixel) || image->depth > bpp)
		return cfb_imageblit( info, image );

	// For clipping
	vxres = S1D_DISPLAY_WIDTH;
	vyres = S1D_DISPLAY_HEIGHT;
	x2 = image->dx + image->width;
	y2 = image->dy + image->height;
	dx = image->dx > 0 ? image->dx : 0;
	dy = image->dy > 0 ? image->dy : 0;
	x2 = x2 < vxres ? x2 : vxres;
	y2 = y2 < vyres ? y2 : vyres;
	width  = x2 - dx;
	height = y2 - dy;

	// Calculate start address
	dstAddr = dx * bpp/8 + dy * STRIDE;

	//Calculate phase
	/*phase = (u32)image->data & 1;*/

	// Wait for any pending blits to finish ...
	if ( !s1d13xxxfb_WaitForBltEnd() )
		return cfb_imageblit( info, image );

	// Select Source and WriteBLT
	s1d13xxxfb_write_reg(REG8008_BLT_COMMAND_ROP, 0x000C0000);

	// Program phase to 0 as linux 2.6 provides byte size bitmap codes (fb_image.data)
	// and the higher depth colors are contained in fg_color/bg_color vars
	//s1d13xxxfb_write_reg(REG800C_BLT_SOURCE_START_ADDR, phase);
	s1d13xxxfb_write_reg(REG800C_BLT_SOURCE_START_ADDR, 0);

	s1d13xxxfb_write_reg(REG8014_BLT_MEMORY_ADDR_OFFSET, STRIDE/2);

	s1d13xxxfb_write_reg(REG8010_BLT_DEST_START_ADDR, dstAddr);

	// Set Width and Height
	s1d13xxxfb_write_reg(REG8018_BLT_WIDTH, width-1);
	s1d13xxxfb_write_reg(REG801C_BLT_HEIGHT, height-1);

	// Engage the blt engine.
	s1d13xxxfb_write_reg(REG8000_BLT_CONTROL, BitBLTCtrlReg | 0x00000001);

	// Wait for blit operation to start
	while ((s1d13xxxfb_read_reg(REG8004_BLT_STATUS) & 0x00000001) == 0) {}

	dst = BltDataAddr;

	// calculate the number of 16 bit words per one blt line
	switch (bpp) {
        case 8:   // 8 bpp
		nTotalWords = ((width + 1 /*+ phase*/) / 2) * height;
		break;
        case 16:  // 16 bpp
		nTotalWords = width * height;
		break;
	}

	if (image->depth == 1)
		s1d13xxxfb_lowblit (info, image,bpp, nTotalWords, dst);
	else
		s1d13xxxfb_colorblit (info, image,bpp, nTotalWords, dst);
}

#endif  //CONFIG_FBCON_EPSON_2D
